import h5py 
import numpy as np
import pandas as pd

import os
import json
import shutil
import argparse

import warnings

from multiprocessing import Pool

from scripts import *

import warnings
warnings.filterwarnings("ignore")

pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', None)
#pd.set_option('display.max_colwidth', None)

def parse_line(command_line):
    parser = argparse.ArgumentParser(description='')
    parser.add_argument("-d", "--dir", help="work direktory", type=str, required=True)
    parser.add_argument("-f", "--simfolder", help="folder to save simulation results", type=str, required=True)
    parser.add_argument("-c", "--coords", help="chosen coordinates in chrA:0,000,000-1,111,111 format", type=str, required=True)
    parser.add_argument("-s", "--singlestrand", help="Make simulation on single strand only", action='store_true')
    parser.add_argument("-n", "--nproc", help="number of threads", type=int, default=1)
    parser.add_argument("-b", "--simbatch", help="how many parallel calculaiotns to store in DRAM (batch_size = {nproc} * {simbatch})", type=int, default=1)
    parser.add_argument("-t", "--steps", help="number of steps per thread", type=int, default=int(2e2))
    parser.add_argument("-w", "--stepswosaving", help="number of steps without saving (total_steps = {steps} * {steps_wo_saving}", type=int, default=int(4.5e2))
    parser.add_argument("-r", "--restarts", help="number of simulations after initiating", type=int, default=1)
    parser.add_argument("-i", "--initsteps", help="number of steps to initiate stable stranscription (default value calculates from max chosen gene length)", type=int, default=None)
    parser.add_argument("-o", "--outname", help="interaction heatmap image outname", type=str, default='simulation.png')

    args = parser.parse_args(command_line)

    return (os.path.realpath(args.dir),
            args.simfolder,  
            args.coords, 
            args.singlestrand, 
            args.nproc, 
            args.nproc * args.simbatch,
            args.steps,
            args.stepswosaving,
            args.restarts,
            args.initsteps,
            args.outname,)


def main(command_line=None):
    (wdir,
     folder_name,
     chosen_slice_str,
     double_strands,
     nproc,
     batch_to_save,
     steps,
     steps_wo_saving,
     restart_simulation,
     initiate_steps,
     out_name) = parse_line(command_line)

    genes_df = gtf_reader(f'{wdir}/valid_transcripts.gtf', transcripts_only=True)
    full_df =  pd.read_csv(f'{wdir}/full.tsv', sep='\t')

    mask = (genes_df['type'] == 'transcript') & (genes_df['FPKM'] > 0)
    max_fpkm = genes_df['FPKM'][mask].max()
    genes_df['len'] = genes_df['end'] - genes_df['start']
    lib_depth = (genes_df[mask]['cov']*genes_df[mask]['len']).sum()
    min_rnap_sep = 30 # bp
    frag_len = 151
    genes_df.loc[mask, 'rnap_sep'] = genes_df[mask][['len', 'FPKM']].apply(lambda x: (1e9 * x[0]) / (x[1] * lib_depth * 2 * frag_len), axis=1)

    args = {
        'BEADSIZE': 45, # bp
        'RNAP_VEL': 100, # bp/s
        'COND_VEL': 1e3, # bp/s
        'RNAP_REM_COND_PROB': 1e-2,
        'BOUNDTIME': 1e0,
        'COND_STEP_OVER_RNAP_PROB': 1e-2,
        'Diffusion_prob': 1e-4,
        'Cohesin_separator': 25e3,
    }
    args['TIMESTEP'] = args['BEADSIZE'] / args['COND_VEL'] # sec, extruder vel ~1 kbp/s and one step ~45 bp
    args['LIFETIME'] = 2e3 / args['TIMESTEP']
    args['LIFETIME_STALLED'] = args['LIFETIME']
    args['ave_loop_size'] = 250e3
    args['steps_wo_saving'] = steps_wo_saving
    args['steps'] = steps

    chosen_slice = slice_str_parser(chosen_slice_str)

    N = chosen_slice[2]//args['BEADSIZE'] - chosen_slice[1]//args['BEADSIZE']
    if double_strands: N = N*2
    args['N'] = N

    mask = (genes_df['chrom'] == chosen_slice[0]) & (genes_df['start'] > chosen_slice[1]) & (genes_df['end'] < chosen_slice[2]) & \
            (genes_df['type'] == 'transcript') & (genes_df['FPKM'] > 0.0)
    chosen_genes = genes_df[mask]

    chosen_genes_beads = genes_to_beads_converter(chosen_genes, full_df, args)

    N_coh = int(np.round(np.round((N/2)*args['BEADSIZE']/args['Cohesin_separator'])))
    maxcondesins = max(int(np.round((N)*args['BEADSIZE']/(args['ave_loop_size']))), 2)
    args['N_coh'] = N_coh
    args['maxcondesins'] = maxcondesins

    initiate_steps = initiate_steps or int(((chosen_genes['len'].max()+1) / args['RNAP_VEL']) / args['TIMESTEP'])
    print(f'initiate_steps {initiate_steps}')

    if os.path.exists(folder_name):
        shutil.rmtree(folder_name)
    os.mkdir(folder_name)

    with open(f'{folder_name}/info', 'w') as file:
        file.write(chosen_slice_str)
    json.dump(args, open(f'{folder_name}/info.json', 'w'), indent = 4)


    max_steps = int(steps * restart_simulation)
    assert max_steps > 0
    # assert (trajectory_len // steps_wo_saving) % steps == 0
    assert max_steps % steps == 0
    max_rnaps = sum([abs(x[0]-x[2])+1 for x in chosen_genes_beads])

    with h5py.File(f'{folder_name}/RNAp_pos_brists.h5', mode='w') as RNApfile, \
        h5py.File(f'{folder_name}/Cohesins_positions.h5', mode='w') as COHfile, \
        h5py.File(f'{folder_name}/Condensins_positions.h5', mode='w') as CONDfile:
    
        RNApfile.attrs["N"] = args['N']
        CONDfile.attrs["N"] = args['N']
        COHfile.attrs["N"] = args['N']
        
        rnap_pos_dset = RNApfile.create_dataset("positions",
                                        shape=(max_steps,
                                                max_rnaps),
                                                dtype=np.int32, 
                                                compression="gzip")
        brist_dset = RNApfile.create_dataset("bristles",
                                            shape=(max_steps,
                                                    max_rnaps),
                                                    dtype=np.int32, 
                                                    compression="gzip")
        
        cond_dset = CONDfile.create_dataset("positions",
                                        shape=(max_steps,
                                                maxcondesins, 2),
                                                dtype=np.int32, 
                                                compression="gzip")
        coh_dset = COHfile.create_dataset("positions",
                                        shape=(max_steps,
                                                N_coh, 2),
                                                dtype=np.int32, 
                                                compression="gzip")
            
        rnaps = []
        occupied_rnap = [[0,0] for _ in range(N)]

        condensins = []
        cohesins = []
        occupied = np.zeros(N)
        occupied[0] = 10
        occupied[-1] = 10
        occupied[N//2-1] = 10
        occupied[N//2] = 10

        for i in range(N_coh): 
            loadOne_sisters(cohesins, occupied, args)

        for _ in range(initiate_steps):
            make_one_step(rnaps, occupied_rnap, chosen_genes_beads, condensins, occupied, maxcondesins, cohesins)
        
        params = (rnaps, occupied_rnap, occupied, cohesins, condensins)
        

        for k in range(restart_simulation//batch_to_save):
            pool = Pool(processes = nproc)
            all_stats = pool.starmap(run_one_simulation, params_generator(params, batch_to_save))
            pool.close()

            for i, (cur_rnap_pos, cur_brists, cur_cond_pos, cur_coh_pos) in enumerate(all_stats):
                    rnap_pos_dset[(k*batch_to_save + i)*steps : (k*batch_to_save + i+1)*steps] = np.array(cur_rnap_pos)
                    brist_dset[(k*batch_to_save + i)*steps : (k*batch_to_save + i+1)*steps] = np.array(cur_brists)
                    cond_dset[(k*batch_to_save + i)*steps : (k*batch_to_save + i+1)*steps] = np.array(cur_cond_pos)
                    coh_dset[(k*batch_to_save + i)*steps : (k*batch_to_save + i+1)*steps] = np.array(cur_coh_pos)
            
            del(all_stats)


    draw_interaction_heatmap(
        args['BEADSIZE'],
        path_to_coh_h5 = f'{folder_name}/Cohesins_positions.h5',
        path_to_cond_h5 = f'{folder_name}/Condensins_positions.h5',
        out_name = out_name,
        first_frame = 0,
        n_frames = None,
        save_each_frame = 1,
        out_resolution = int(4e3),
        interaction_presence_only = True,
        log_scale = True,
        figsize = (10,10),
    #     sector = [[418e3, 475e3], [418e3, 475e3]],
        chunk = int(1e5),
    )



if __name__ == "__main__":
    main()